package cmd

import (
	"fmt"
	"io"
	"os"
	"path"

	"github.com/spf13/cobra"
	"xbus.io/go-xbus/v3/api"
)

var pipelineGetAllCmd = &cobra.Command{
	Use:   "get-all <name or id> [version]",
	Short: "Get all the pipeline definitions",
	Long:  "Get all the pipeline definitions and write them to one or multiple files",
	Run: func(cmd *cobra.Command, args []string) {
		if len(args) != 0 {
			usage(cmd)
		}

		var (
			output                 = cmd.Flag("output").Value.String()
			multipleFiles          = cmd.Flag("multiple-files").Value.String() == "true"
			multipleFilesNoVersion = cmd.Flag("multiple-files-no-version").Value.String() == "true"
		)

		c, err := GetClient(false)
		if err != nil {
			Logger.Errorf("%s", err)
			os.Exit(-1)
		}

		filter, err := parsePipelineFilterFlags(cmd)
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}

		pipelines, _, err := filterPipelineList(c, filter)
		if err != nil {
			Logger.Errorf("%s", err)
			return
		}

		var out io.Writer
		if !multipleFiles {
			if output == "-" {
				out = os.Stdout
			} else {
				f, err := os.Create(output)
				if err != nil {
					Logger.Errorf("Error opening output file: %s", err)
					os.Exit(-1)
				}
				defer func() {
					if err := f.Close(); err != nil {
						Logger.Errorf("Error saving pipeline: %s", err)
						os.Exit(-1)
					}
				}()
				out = f
			}
			if _, err := out.Write([]byte("pipelines:\n")); err != nil {
				Logger.Errorf("Error opening output file: %s", err)
				os.Exit(-1)
			}
		}
		if multipleFilesNoVersion {
			var newSet = map[string]api.PipelineInfo{}

			for _, pipeline := range pipelines {
				switch pipeline.Status {
				case api.PipelineInfo_DRAFT:
					newSet[pipeline.Name] = pipeline
				case api.PipelineInfo_ACTIVE:
					if _, ok := newSet[pipeline.Name]; !ok {
						newSet[pipeline.Name] = pipeline
					}
				case api.PipelineInfo_INACTIVE:
				}
			}
			pipelines = pipelines[0:0]
			for _, pipeline := range newSet {
				pipelines = append(pipelines, pipeline)
			}
		}

		for _, pipeline := range pipelines {
			var (
				f     io.WriteCloser
				fname string
				pout  io.Writer
			)
			if multipleFiles {
				fname = fmt.Sprintf("%s-%s.yaml", pipeline.Name, pipeline.Version)
				if multipleFilesNoVersion {
					fname = fmt.Sprintf("%s.yaml", pipeline.Name)
				}
				if fname != "-" {
					fname = path.Join(output, fname)
				}
				var err error
				f, err = os.Create(fname)
				if err != nil {
					Logger.Errorf("Error opening output file: %s", err)
					os.Exit(-1)
				}
				if _, err := fmt.Fprintf(
					f, "version: \"%s\"\nstatus: %s\n", pipeline.Version, pipeline.Status,
				); err != nil {
					Logger.Errorf("Error writing file: %s", err)
					if err := f.Close(); err != nil {
						Logger.Errorf("Error closing file: %s", err)
					}
					os.Exit(-1)
				}
				pout = f
			} else {
				version := pipeline.Version
				if version == "" {
					version = `""`
				}
				if _, err := out.Write([]byte(
					fmt.Sprintf(
						"  %s:\n    %s:\n      status: %s\n",
						pipeline.Name, version, pipeline.Status),
				)); err != nil {
					Logger.Errorf("Error writing file: %s", err)
					os.Exit(-1)
				}
				pout = IndentWriter{[]byte("      "), out, false}
			}
			if err := pipelineGet(c, pipeline.Name, pipeline.Version, pout); err != nil {
				Logger.Errorf("Error getting pipeline: %s", err)
				os.Exit(-1)
			}
			if multipleFiles {
				if err := f.Close(); err != nil {
					Logger.Errorf("Error closing file: %s", err)
					os.Exit(-1)
				}
				Logger.Noticef("Wrote %s", fname)
			}
		}
	},
}

func init() {
	pipelineGetAllCmd.Flags().Bool("multiple-files", false, "Write the pipelines in individual files")
	pipelineGetAllCmd.Flags().Bool("multiple-files-no-version", false, "Don't include the pipeline version in the file name")
	pipelineGetAllCmd.Flags().StringP("output", "o", "-", "A file or a directory, defaults to '-' or '.' depending on the mode")
	addPipelineFilterFlags(pipelineGetAllCmd.Flags(), true)
	pipelineCmd.AddCommand(pipelineGetAllCmd)
}
