// Copyright 2016 Orus.io. All rights reserved.

package director

import (
	"context"
	"fmt"
	"strings"
	"sync"
	"time"

	nats "github.com/nats-io/go-nats"
	nrpc "github.com/nats-rpc/gogo-nrpc"
	"github.com/prometheus/client_golang/prometheus"

	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/tools"
	"xbus.io/xbus/process"
	"xbus.io/xbus/services"
	svc_tools "xbus.io/xbus/services/tools"
	"xbus.io/xbus/storage"
)

func min(a, b uint) uint {
	if a < b {
		return a
	}
	return b
}

const (
	// DirectorQueue is the name we use to queueSub to a queue with the director
	DirectorQueue = "director_queue"
)

// Options is used to create a Service
type Options struct {
	services.Options

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	LogStorage      storage.LogStorage
	ProcessManager  *process.Manager
	JobStorage      storage.JobStorage
}

// NewService returns a new Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService:    services.NewBaseService("director", o.Options),
		Opts:           o,
		processManager: o.ProcessManager,
		running:        make(map[api.UUID][]*Job),
		envWatcher:     NewEnvelopeWatcher(),
	}
}

// New returns a new Service
func New(options Options) *Service {
	return options.NewService()
}

// Job ...
type Job struct {
	storage.Job
	timeoutCount uint

	// If currently running, cancel will stop the job
	cancel func()
}

// Service is the director service implementation
// The director service is responsible for calling the 'Process' function
// on the actors.
// It monitors the ProcessNodeReadyEvent events and schedule jobs
// When a actor is back online or a when a process is resumed, the corresponding
// jobs are resheduled
// The director service acts alone a is not (yet) able to cooperate with other
// director instances.
type Service struct {
	services.BaseService
	Opts Options

	processManager *process.Manager

	api             *api.Client
	directorHandler *api.DirectorHandler
	actorMap        *svc_tools.ActorMap
	jobs            chan Job
	pushback        chan *Job

	envWatcher *EnvelopeWatcher

	ctx    context.Context
	cancel func()

	done chan struct{}

	// Jobs for which 'Process' is being called so they can be canceled
	running     map[api.UUID][]*Job
	runningLock sync.Mutex
}

func (s *Service) mainloop() {
	defer s.SafeTaskDone()

	actorQueues := make(map[api.UUID]*JobQueue)

	defer func() {
		for _, queue := range actorQueues {
			queue.Close()
		}
	}()

	setOnline := func(id api.UUID) {
		// init the queue if needed
		queue, ok := actorQueues[id]
		if !ok {
			queue = newJobQueue(
				s.Opts.JobStorage, id, 1024,
			)
			actorQueues[id] = queue
		}

		// add a worker on the queue
		// TODO for parallel-able actors, start more routine
		s.TaskAdd()
		go s.worker(queue.C())
	}

	for _, id := range s.actorMap.ListOnline() {
		setOnline(id)
	}

	for {
		select {
		case job := <-s.jobs:
			queue, ok := actorQueues[job.Node.ActorID]
			// If a queue exists, add the job to it
			if ok {
				queue.Add(&job)
			}
		case event := <-s.actorMap.Events:
			if event.Online {
				setOnline(event.ID)
			} else {
				// drop the queue worker and mark the queue for deletion
				if queue, ok := actorQueues[event.ID]; ok {
					queue.SendNilToOut()
					// TODO mark the queue for deletion
				}
			}
		case job := <-s.pushback:
			queue, ok := actorQueues[job.Node.ActorID]

			// If a queue exists, re-add the job to it in first position
			if ok {
				queue.PushBack(job)
			}

		case <-s.done:
			return
		}
	}
}

// Startup initialize all the subscriptions
func (s *Service) Startup() error {
	s.ctx, s.cancel = context.WithCancel(context.Background())
	s.directorHandler = api.NewDirectorHandler(
		s.ctx, s.Conn, directorServer{s},
	)
	if err := s.SubscribeAll(
		s.HandlerQueueSubscribe(s.directorHandler, DirectorQueue),
	); err != nil {
		return err
	}

	s.done = make(chan struct{})
	s.jobs = make(chan Job)
	s.pushback = make(chan *Job)

	s.api = api.NewClient(s.Conn)

	sub, err := s.api.CoreEvents.ProcessNodeReadySubscribe(
		s.handleProcessNodeReady)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	sub, err = s.api.CoreEvents.ProcessSubscribe(s.handleProcess)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	if m, err := svc_tools.NewActorMap(s.ctx, s.api, s.Log); err == nil {
		s.actorMap = m
	} else {
		return err
	}

	sub, err = s.api.CoreEvents.EnvelopeSubscribe(s.envWatcher.PushEvent)
	if err != nil {
		return err
	}
	s.AddSubscription(sub)

	s.TaskAdd()
	go s.mainloop()

	return nil
}

// Shutdown the service
func (s *Service) Shutdown() error {
	s.actorMap.Close()
	close(s.done)
	t := time.AfterFunc(5*time.Second, func() {
		s.Log.Info().Msg("Forcing shutdown of current calls")
		s.cancel()
	})
	err := s.BaseService.Shutdown()
	t.Stop()
	close(s.jobs)
	s.jobs = nil
	return err
}

type directorServer struct {
	*Service
}

func (s directorServer) Log(
	ctx context.Context, actorSID string, req api.ActorLogRequest,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	if req.EnvelopeID != nil && !req.EnvelopeID.IsZero() {
		_, err = s.Opts.ProcessManager.Get(req.Context.ProcessID)
		if err != nil {
			return
		}
	}

	var (
		envID, processID api.UUID
		nodeID           string
	)
	if req.EnvelopeID != nil {
		envID = *req.EnvelopeID
	}
	if ctx := req.Context; ctx != nil {
		nodeID = ctx.NodeID
		processID = ctx.ProcessID
	}
	for _, msg := range req.Messages {
		s.Opts.LogStorage.Log(api.LogEntry{
			EnvelopeID: envID,
			ActorID:    actorID,
			NodeID:     nodeID,
			ProcessID:  processID,
			Message:    msg,
		})
	}
	return
}

func (s directorServer) ProcessingEnd(
	ctx context.Context, actorSID string, req api.ActorProcessingState,
) (err error) {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return
	}

	p, err := s.Opts.ProcessManager.Get(req.Context.ProcessID)
	if err != nil {
		return
	}

	if req.Status != api.ActorProcessingState_SUCCESS &&
		req.Status != api.ActorProcessingState_ERROR {
		return fmt.Errorf("Invalid status: %s", req.Status)
	}

	for _, msg := range req.Messages {
		s.Opts.LogStorage.Log(api.LogEntry{
			ActorID:   actorID,
			NodeID:    req.Context.NodeID,
			ProcessID: req.Context.ProcessID,
			Message:   msg,
		})
	}

	s.Opts.JobStorage.SetState(storage.JobID{
		ProcessID: req.Context.ProcessID,
		Node: storage.ProcessNode{
			ActorID: actorID,
			NodeID:  req.Context.NodeID,
		},
	}, storage.JobState{
		Detached: true,
		Status:   req.Status,
	})
	p.ActorProcessingEnd(actorID, req.Context.NodeID, req.Status, s.Opts.Logger)
	return
}

func (s directorServer) GetRunningActorRequestList(
	ctx context.Context, actorSID string,
	send func(api.ActorProcessRequest),
) error {
	actorID, err := api.UUIDFromString(actorSID)
	if err != nil {
		return err
	}
	jobList, err := s.Opts.JobStorage.ListDetachedJobByActor(actorID)
	if err != nil {
		return err
	}
	defer jobList.Close()
	for jobList.Next() && ctx.Err() == nil {
		job, err := jobList.Get()
		if err != nil {
			return err
		}
		req := api.ActorProcessRequest{
			Context: &api.ProcessingContext{
				ProcessID: job.ProcessID,
				NodeID:    job.Node.NodeID,
			},
		}

		for _, input := range s.processManager.ProcessStorage.GetInputs(
			job.ProcessID, job.Node,
		) {
			envelope, pos, err := s.Opts.EnvelopeStorage.ReadEnvelope(
				input.EnvelopeID, storage.EnvelopeStartPos(), 100*1024)
			if err != nil {
				s.Service.Log.Error().
					Str("envelope-id", input.EnvelopeID.String()).
					Msg("could not read envelope")
				return err
			}
			req.Inputs = append(req.Inputs, &api.ActorProcessRequest_Input{
				Name:     input.Name,
				Envelope: &envelope,
				Position: storage.EncodePosition(envelope.ID, pos),
			},
			)
		}
		if ctx.Err() == nil {
			send(req)
		}
	}
	return ctx.Err()
}

func (s *Service) handleProcessNodeReady(event api.ProcessNodeReadyEvent) {
	sJob := storage.Job{
		JobID: storage.JobID{
			ProcessID: event.ProcessID,
			Node: storage.ProcessNode{
				ActorID: event.ActorID,
				NodeID:  event.NodeID,
			},
		},
		JobState: storage.JobState{
			CreatedAt: time.Now(),
		},
	}
	s.Opts.JobStorage.NewJob(sJob.JobID, sJob.JobState)

	s.jobs <- Job{Job: sJob}
}

func (s *Service) handleProcess(event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "status" {
		if event.ProcessOldStatus == api.Process_PAUSED &&
			event.Process.Status == api.Process_RUNNING {

			for _, j := range s.Opts.JobStorage.ListJobByProcess(event.Process.ID) {
				s.Log.Info().Str("node", j.Node.String()).
					Msg("rescheduling processings")
				s.jobs <- Job{Job: j}
			}
		} else if event.Process.Status == api.Process_ERROR {
			// we need to cancel ongoing processing if any
			s.cancelProcessJobs(event.Process.ID)
		}
	}
}

func (s *Service) worker(jobs <-chan *Job) {
	defer s.SafeTaskDone()

	for {
		select {
		case job := <-jobs:
			if job == nil {
				s.Log.Debug().Msg("Worker: Exiting")
				return
			}
			s.handleJob(job)
		case <-s.ctx.Done():
			return
		}
	}
}

func (s *Service) setJobRunning(job *Job) (context.Context, func()) {
	s.runningLock.Lock()
	defer s.runningLock.Unlock()

	ctx, cancel := context.WithCancel(s.ctx)
	job.cancel = cancel

	processJobList := s.running[job.ProcessID]
	s.running[job.ProcessID] = append(processJobList, job)

	return ctx, cancel
}

func (s *Service) unsetJobRunning(job *Job) {
	s.runningLock.Lock()
	defer s.runningLock.Unlock()

	processJobList := s.running[job.ProcessID]
	for i, j := range processJobList {
		if j == job {
			processJobList = append(
				processJobList[:i], processJobList[i+1:]...)
			break
		}
	}
	if len(processJobList) == 0 {
		delete(s.running, job.ProcessID)
	} else {
		s.running[job.ProcessID] = processJobList
	}
}

func (s *Service) cancelProcessJobs(id api.UUID) {
	s.runningLock.Lock()
	defer s.runningLock.Unlock()

	processJobList := s.running[id]
	for _, job := range processJobList {
		job.cancel()
	}
}

func (s *Service) handleJob(job *Job) {
	s.Log.Debug().Str("job-id", job.JobID.String()).Msg("handling job")

	ctx, cancel := s.setJobRunning(job)
	defer s.unsetJobRunning(job)

	actor, ok := s.Opts.ActorStorage.Get(job.Node.ActorID)
	if !ok {
		s.Log.Error().Str("agent-id", job.Node.ActorID.String()).
			Msg("no such agent")
		return
	}
	if !actor.Online {
		s.Log.Info().Str("agent", actor.Name).
			Msg("agent is offline.")
		s.pushback <- job
		return
	}
	nc := tools.NewServerToClientConn(s.Conn, actor.AccountID.String())
	agent := api.NewActorAgentClient(nc, job.Node.ActorID.String())
	req := api.ActorProcessRequest{
		Context: &api.ProcessingContext{
			ProcessID: job.ProcessID,
			NodeID:    job.Node.NodeID,
		},
	}

	// XXX The following lines loads the process and pipeline only to get
	// the pipeline name and version for metrics. It may slow down things.
	process, ok := s.processManager.ProcessStorage.GetProcess(job.ProcessID)
	if !ok {
		s.Log.Error().
			Str("process-id", job.ProcessID.String()).
			Msg("unknown process id")
		return
	}
	pipeline, ok := s.processManager.PipelineManager.GetPipeline(process.PipelineID)
	if !ok {
		s.Log.Error().
			Str("pipeline-id", process.PipelineID.String()).
			Msg("unknown pipeline id")
		return
	}

	if s.processManager.ProcessStorage.GetState(job.ProcessID).Status > api.Process_RUNNING {
		return
	}
	var envelopeIDs []api.UUID
	for _, input := range s.processManager.ProcessStorage.GetInputs(
		job.ProcessID, job.Node,
	) {
		envelope, pos, err := s.Opts.EnvelopeStorage.ReadEnvelope(
			input.EnvelopeID, storage.EnvelopeStartPos(), 100*1024)
		if err != nil {
			s.Log.Err(err).
				Str("envelope-id", input.EnvelopeID.String()).
				Msg("could not read envelope")
			return
		}
		req.Inputs = append(req.Inputs, &api.ActorProcessRequest_Input{
			Name:     input.Name,
			Envelope: &envelope,
			Position: storage.EncodePosition(envelope.ID, pos),
		},
		)
		envelopeIDs = append(envelopeIDs, envelope.ID)
	}

	job.Status = api.ActorProcessingState_RUNNING
	job.EnvelopeIDs = envelopeIDs
	s.Opts.JobStorage.SetState(job.JobID, job.JobState)

	// timeoutCount only apply to the current set of attempts
	job.timeoutCount = 0

	watchSub := s.envWatcher.Watch(func(event api.EnvelopeEvent) {
		switch event.Status {
		case api.EnvelopeEvent_STALLED, api.EnvelopeEvent_ERROR:
			// Cancel any current call
			cancel()

			// Cancel the whole process
			process, err := s.Opts.ProcessManager.Get(job.JobID.ProcessID)
			if err != nil {
				s.Log.Err(err).
					Str("process-id", job.JobID.ProcessID.String()).
					Msg("error getting process")
			}

			if err := process.Cancel(
				fmt.Sprintf("Envelope %s status is %s", event.EnvelopeID, event.Status),
			); err != nil {
				s.Log.Err(err).
					Str("process-id", job.JobID.ProcessID.String()).
					Msg("error canceling process")
			}

		// TODO Mark all the unfinished envelopes as error so they cannot receive
		// more fragments (not here actually, but in the process cancellation handling)
		default:
		}

	}, envelopeIDs...)
	defer watchSub.Unsubscribe()

	for {
		s.Log.Debug().
			Str("actor", actor.Name).
			Uint("attempt", job.AttemptCount).
			Msg("calling Process")

		processStart := time.Now()
		if job.FirstAttempt.IsZero() {
			job.FirstAttempt = processStart
			s.Opts.JobStorage.SetState(job.JobID, job.JobState)
		}

		err := agent.Process(
			ctx, req,
			func(ctx context.Context, state api.ActorProcessingState) {
				job.Status = state.Status
				s.Log.Debug().
					Interface("jobstate", job.JobState).
					Msg("saving job new state")
				if state.ActorLeaving {
					s.actorMap.SetOnline(job.Node.ActorID, false)
				}
				s.Opts.JobStorage.SetState(job.JobID, job.JobState)
				switch state.Status {
				case api.ActorProcessingState_RUNNING:
				default:
					if err := (directorServer{s}.ProcessingEnd(ctx, job.Node.ActorID.String(), state)); err != nil {
						s.Log.Err(err).Msg("error calling ProcessingEnd")
					}
				}
			},
		)

		processEnd := time.Now()

		job.AttemptCount++

		s.Opts.JobStorage.SetState(job.JobID, job.JobState)

		if err == nil {
			status := strings.ToLower(job.Status.String())

			if job.Status == api.ActorProcessingState_RUNNING {
				// the job is detached, and will report its status asynchronously
				// with the Director.ProcessingEnd API entry point
				job.Detached = true
				s.Opts.JobStorage.SetState(job.JobID, job.JobState)

				status = "detached"
			}
			observeJobDurations(
				actor.Name, pipeline.Name, pipeline.Version, job.Node.NodeID,
				status, job.CreatedAt, job.FirstAttempt, processStart, processEnd)

			if err := s.Opts.ActorStorage.JustSeen(job.Node.ActorID); err != nil {
				s.Log.Err(err).
					Str("actor", job.Node.ActorID.String()).
					Msg("error calling JustSeen")
			}
			return
		}

		s.Log.Info().Err(err).Str("actor", actor.Name).Msg("Process failed")
		if err == nats.ErrTimeout {
			job.timeoutCount++
			s.Log.Error().
				Str("actor", actor.Name).
				Uint("count", job.timeoutCount).
				Msg("agent unresponsive")
			if job.timeoutCount == 1 {
				s.Log.Info().
					Str("actor", actor.Name).
					Msg("marking actor unresponsive")
				if err := s.Opts.ActorStorage.MarkUnresponsive(job.Node.ActorID, job.FirstAttempt); err != nil {
					s.Log.Err(err).Msg("failed to mark unresponsive")
				}
			} else if job.timeoutCount >= 5 {
				s.Log.Error().
					Str("actor", actor.Name).
					Msg("marking actor offline after 5 timeouts on process attempts")
				if err := s.Opts.ActorStorage.SetOnline(job.Node.ActorID, false); err != nil {
					s.Log.Err(err).Msg("error marking offline")
				}
				s.pushback <- job
				return
			}
			s.Log.Debug().
				Dur("duration", time.Duration(job.timeoutCount)*250).
				Msg("Sleeping")
		} else if nrpcErr, ok := err.(*nrpc.Error); ok && nrpcErr.GetType() == nrpc.Error_SERVERTOOBUSY {
			// Back to pending, it case it is all restarted
			job.Status = api.ActorProcessingState_PENDING
			s.Opts.JobStorage.SetState(job.JobID, job.JobState)
			job.timeoutCount = 0
			delay := (1 << min(job.AttemptCount, 11)) * time.Second
			s.Log.Info().
				Str("actor", actor.Name).
				Dur("duration", delay).
				Str("reason", nrpcErr.Message).
				Msg("Agent cannot handle Process() now, wait some time before retrying.")
			select {
			case <-ctx.Done():
				return
			case <-time.After(delay):
			}
		} else {
			s.Log.Err(err).
				Str("actor", actor.Name).
				Uint("attempt", job.AttemptCount).
				Msg("Error calling agent Process")
			observeJobDurations(
				actor.Name, pipeline.Name, pipeline.Version, job.Node.NodeID,
				"unexpected error", job.CreatedAt, job.FirstAttempt, processStart, processEnd)
			return
		}

	}
}

var (
	jobWaitTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_wait_time_seconds",
			Help:       "The actor job wait in queue before effective process, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)

	jobBusyTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_busy_time_seconds",
			Help:       "The actor busy time before effective process, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)

	jobProcessTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_process_time_seconds",
			Help:       "The actor job process time, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)

	jobTotalTime = prometheus.NewSummaryVec(
		prometheus.SummaryOpts{
			Name:       "xbus_service_director_job_total_time_seconds",
			Help:       "The actor job wait and execution time, measured from the director service.",
			Objectives: map[float64]float64{0.9: 0.01, 0.95: 0.01, 0.99: 0.001},
			ConstLabels: map[string]string{
				"service": "director",
			},
		},
		[]string{"actor", "pipeline", "pipeline_version", "node", "status"},
	)
)

func observeJobDurations(actor, pipeline, version, node, status string, jobCreatedAt, firstattempt, processStart, processEnd time.Time) {

	waitDuration := firstattempt.Sub(jobCreatedAt)
	busyDuration := processStart.Sub(firstattempt)
	processDuration := processEnd.Sub(processStart)
	totalDuration := waitDuration + busyDuration + processDuration

	jobWaitTime.
		WithLabelValues(actor, pipeline, version, node, status).
		Observe(waitDuration.Seconds())

	jobBusyTime.
		WithLabelValues(actor, pipeline, version, node, status).
		Observe(busyDuration.Seconds())

	jobProcessTime.
		WithLabelValues(actor, pipeline, version, node, status).
		Observe(processDuration.Seconds())

	jobTotalTime.
		WithLabelValues(actor, pipeline, version, node, status).
		Observe(totalDuration.Seconds())
}

func init() {
	prometheus.MustRegister(jobWaitTime)
	prometheus.MustRegister(jobBusyTime)
	prometheus.MustRegister(jobProcessTime)
	prometheus.MustRegister(jobTotalTime)
}
