package metrics

import (
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api/control"
)

// NewEnvelopeStatCollector creates a new EnvelopeStatCollector
func NewEnvelopeStatCollector(
	control *control.Client, delayHint time.Duration, log zerolog.Logger,
) *EnvelopeStatCollector {
	c := EnvelopeStatCollector{
		control: control,
		log:     log,

		storageCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "count"),
			"Number of items in the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "envelope"},
		),
		storageSizeDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "size"),
			"Total size of the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "envelope"},
		),
		queryChan: make(chan envelopeStatEntry),
		maxAge:    delayHint * 2,
		minDelay:  delayHint - 2*time.Second,
	}
	go c.run()
	return &c
}

type envelopeStatEntry struct {
	stat      control.StorageStat
	timestamp time.Time
	duration  time.Duration
}

// EnvelopeStatCollector implements the prometheus.Collector interface
type EnvelopeStatCollector struct {
	control *control.Client
	log     zerolog.Logger

	storageCountDesc *prometheus.Desc
	storageSizeDesc  *prometheus.Desc

	maxAge   time.Duration
	minDelay time.Duration

	queryChan chan envelopeStatEntry
	stop      chan struct{}

	lastEntry *envelopeStatEntry
}

// Describe implements the prometheus.Collector interface.
func (c EnvelopeStatCollector) Describe(ch chan<- *prometheus.Desc) {
	ch <- c.storageCountDesc
	ch <- c.storageSizeDesc
}

// Collect implements the prometheus.Collector interface.
func (c *EnvelopeStatCollector) Collect(ch chan<- prometheus.Metric) {
	timeout := time.After(100 * time.Millisecond)
	for {
		select {
		case <-timeout:
			if c.lastEntry != nil && time.Since(c.lastEntry.timestamp) <= c.maxAge {
				c.collect(c.lastEntry.stat, ch)
			} else {
				c.log.Warn().
					Str("collector", "envelopestat").
					Msg("no metrics data available. The metrics delay hint is probably too low. See https://docs.xbus.io/administration/metrics.html for more information")
			}
			return
		case r := <-c.queryChan:
			if time.Since(r.timestamp) > c.maxAge {
				continue
			}
			c.collect(r.stat, ch)
			c.lastEntry = &r
			return
		}
	}
}

func (c *EnvelopeStatCollector) collect(stat control.StorageStat, ch chan<- prometheus.Metric) {

	for _, entry := range stat.Entries {
		ch <- prometheus.MustNewConstMetric(
			c.storageCountDesc,
			prometheus.GaugeValue,
			float64(entry.Count),
			entry.Name,
			entry.Implementation,
		)

		ch <- prometheus.MustNewConstMetric(
			c.storageSizeDesc,
			prometheus.GaugeValue,
			float64(entry.EstimatedSize),
			entry.Name,
			entry.Implementation,
		)
	}
}

func (c *EnvelopeStatCollector) run() {
	var (
		lastTime     time.Time
		lastDuration time.Duration
	)

	for {
		if lastDuration != 0 {
			delay := c.minDelay - lastDuration - time.Since(lastTime)
			if delay > 0 {
				select {
				case <-time.After(delay):
				case <-c.stop:
					return
				}
			}
		}

		var (
			start = time.Now()
		)

		stat, err := c.control.Envelope.GetStorageStat()
		if err != nil {
			c.log.Err(err).Msg("metrics: error reading envelope storage stats")
			return
		}

		lastTime = time.Now()
		lastDuration = lastTime.Sub(start)
		select {
		case <-c.stop:
			return
		case c.queryChan <- envelopeStatEntry{stat, lastTime, lastDuration}:
		}
	}
}
