package metrics

import (
	"context"
	"fmt"
	"time"

	"github.com/prometheus/client_golang/prometheus"
	"github.com/rs/zerolog"
	"xbus.io/go-xbus/v3/api"
	"xbus.io/go-xbus/v3/api/control"
	"xbus.io/xbus/services/tools"
)

// NewSummaryCollector creates a new SummaryCollector
func NewSummaryCollector(
	control *control.Client, actors *tools.ActorCache, pipelines *tools.PipelineCache,
	delayHint time.Duration, log zerolog.Logger,
) *SummaryCollector {
	c := SummaryCollector{
		log:       log,
		control:   control,
		actors:    actors,
		pipelines: pipelines,

		processCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "process", "count"),
			"Number of process in the bus",
			[]string{"pipeline", "pipeline_version", "emitter", "status", "result_acked"},
			nil,
		),

		storageCountDesc: prometheus.NewDesc(
			prometheus.BuildFQName("xbus", "storage", "count"),
			"Number of items in the storage",
			[]string{"name", "implementation"},
			prometheus.Labels{"type": "process"},
		),

		queryChan: make(chan entry),
		maxAge:    delayHint * 2,
		minDelay:  delayHint - 2*time.Second,
	}
	c.activate()
	return &c
}

type entry struct {
	summary   control.ProcessSummary
	timestamp time.Time
	duration  time.Duration
}

// SummaryCollector implements the prometheus.Collector interface
type SummaryCollector struct {
	log zerolog.Logger

	control   *control.Client
	actors    *tools.ActorCache
	pipelines *tools.PipelineCache

	storageCountDesc *prometheus.Desc
	processCountDesc *prometheus.Desc

	maxAge   time.Duration
	minDelay time.Duration

	queryChan chan entry
	stop      chan struct{}

	lastEntry *entry
}

/* activate starts 'run' if needed */
func (c *SummaryCollector) activate() {
	go c.run()
}

func (c *SummaryCollector) run() {
	var (
		lastSummaryTime time.Time
		lastDuration    time.Duration
	)

	for {
		if lastDuration != 0 {
			delay := c.minDelay - lastDuration - time.Since(lastSummaryTime)
			if delay > 0 {
				select {
				case <-time.After(delay):
				case <-c.stop:
					return
				}
			}
		}

		var (
			summary control.ProcessSummary
			start   = time.Now()
		)

		if err := c.control.Process.QuerySummary(
			context.Background(), control.ProcessQueryRequest{},
			func(ctx context.Context, r control.ProcessSummary) {
				summary.Entries = append(summary.Entries, r.Entries...)
			},
		); err != nil {
			return
		}

		lastSummaryTime = time.Now()
		lastDuration = lastSummaryTime.Sub(start)
		select {
		case <-c.stop:
			return
		case c.queryChan <- entry{summary, lastSummaryTime, lastDuration}:
		}
	}
}

// Describe implements the prometheus.Collector interface.
func (c SummaryCollector) Describe(ch chan<- *prometheus.Desc) {
	ch <- c.processCountDesc
}

func (c *SummaryCollector) getActor(id api.UUID) string {
	a, err := c.actors.Get(id)
	if err != nil {
		return id.String()
	}
	return a.Name
}

func (c *SummaryCollector) getPipeline(id api.UUID) (string, string) {
	p, err := c.pipelines.Get(id)
	if err != nil {
		return id.String(), ""
	}
	return p.Name, p.Version
}

func (c *SummaryCollector) collect(summary control.ProcessSummary, ch chan<- prometheus.Metric) {
	var count uint64

	for _, entry := range summary.Entries {
		count += entry.Count
		emitter := c.getActor(entry.EmitterID)
		pipeline, version := c.getPipeline(entry.PipelineID)
		ch <- prometheus.MustNewConstMetric(
			c.processCountDesc,
			prometheus.GaugeValue,
			float64(entry.Count),
			pipeline,
			version,
			emitter,
			entry.Status.String(),
			fmt.Sprintf("%v", entry.ResultAcked),
		)
	}

	ch <- prometheus.MustNewConstMetric(
		c.storageCountDesc,
		prometheus.GaugeValue,
		float64(count),
		"default",
		"default",
	)
}

// Collect implements the prometheus.Collector interface.
func (c *SummaryCollector) Collect(ch chan<- prometheus.Metric) {
	timeout := time.After(100 * time.Millisecond)
	for {
		select {
		case <-timeout:
			if c.lastEntry != nil {
				if time.Since(c.lastEntry.timestamp) <= c.maxAge {
					c.collect(c.lastEntry.summary, ch)
				}
			} else {
				c.log.Warn().
					Str("collector", "process-summary").
					Msg("no metrics data available. The metrics delay hint is probably too low. See https://docs.xbus.io/administration/metrics.html for more information")
			}
			return
		case r := <-c.queryChan:
			if time.Since(r.timestamp) > c.maxAge {
				continue
			}
			c.collect(r.summary, ch)
			c.lastEntry = &r
			return
		}
	}
}
